//! File IO for generated codes.

use std::{collections::HashMap, env, ffi::OsString, fs, io, path, process};

use thiserror::Error;

use crate::types::XmlGitInfo;

const ENVVAR_XML_GIT_DESCRIBE: &str = "XML_GIT_DESCRIBE";
const ENVVAR_XML_GIT_TIMESTAMP: &str = "XML_GIT_TIMESTAMP";
const ENVVAR_XML_GIT_SHA: &str = "XML_GIT_SHA";

#[derive(Debug, Error)]
pub enum XmlGitInfoError {
    #[error("io error while getting xml git info: {msg} {source}")]
    GitCommandSpawnError {
        msg: &'static str,
        #[source]
        source: io::Error,
    },
    #[error("git failed: {msg}: {stderr}")]
    CommandFailed { msg: &'static str, stderr: String },
    #[error("env var {var} is not unicode: {}", s.to_string_lossy())]
    NotUnicodeEnvVar { var: &'static str, s: OsString },
}

impl XmlGitInfoError {
    fn git_command_spawn_error(e: io::Error, msg: &'static str) -> Self {
        XmlGitInfoError::GitCommandSpawnError { msg, source: e }
    }
}

pub(crate) fn get_xml_gitinfo(xml_path: &path::Path) -> Result<XmlGitInfo, XmlGitInfoError> {
    let describe = match env::var(ENVVAR_XML_GIT_DESCRIBE) {
        Ok(describe) => describe,
        Err(env::VarError::NotUnicode(s)) => {
            return Err(XmlGitInfoError::NotUnicodeEnvVar {
                var: ENVVAR_XML_GIT_DESCRIBE,
                s,
            });
        }
        Err(env::VarError::NotPresent) => {
            let out_describe = process::Command::new("git")
                .args(["describe", "--always", "--dirty"])
                .current_dir(xml_path.parent().unwrap())
                .output()
                .map_err(|e| XmlGitInfoError::git_command_spawn_error(e, "git describe"))?;
            let describe = if out_describe.status.success() {
                String::from_utf8_lossy(&out_describe.stdout)
                    .trim()
                    .to_owned()
            } else {
                return Err(XmlGitInfoError::CommandFailed {
                    msg: "git describe",
                    stderr: String::from_utf8_lossy(&out_describe.stderr).into_owned(),
                });
            };
            describe
        }
    };
    let timestamp = match env::var(ENVVAR_XML_GIT_TIMESTAMP) {
        Ok(timestamp) => timestamp,
        Err(env::VarError::NotUnicode(s)) => {
            return Err(XmlGitInfoError::NotUnicodeEnvVar {
                var: ENVVAR_XML_GIT_TIMESTAMP,
                s,
            });
        }
        Err(env::VarError::NotPresent) => {
            let timestamp_out = process::Command::new("git")
                .args(["log", "-1", "--pretty=format:'%cI'"])
                .current_dir(xml_path.parent().unwrap())
                .output()
                .map_err(|e| XmlGitInfoError::git_command_spawn_error(e, "git log (timestamp)"))?;
            let timestamp = if timestamp_out.status.success() {
                String::from_utf8_lossy(&timestamp_out.stdout)
                    .trim()
                    .to_owned()
            } else {
                return Err(XmlGitInfoError::CommandFailed {
                    msg: "git log (timestamp)",
                    stderr: String::from_utf8_lossy(&timestamp_out.stderr).into_owned(),
                });
            };
            timestamp
        }
    };
    let git_sha = match env::var(ENVVAR_XML_GIT_SHA) {
        Ok(sha) => sha,
        Err(env::VarError::NotUnicode(s)) => {
            return Err(XmlGitInfoError::NotUnicodeEnvVar {
                var: ENVVAR_XML_GIT_SHA,
                s,
            });
        }
        Err(env::VarError::NotPresent) => {
            let sha = process::Command::new("git")
                .args(["rev-parse", "HEAD"])
                .current_dir(xml_path.parent().unwrap())
                .output()
                .map_err(|e| XmlGitInfoError::git_command_spawn_error(e, "git rev-parse (sha)"))?;
            let git_sha = if sha.status.success() {
                String::from_utf8_lossy(&sha.stdout).trim().to_owned()
            } else {
                return Err(XmlGitInfoError::CommandFailed {
                    msg: "git rev-parse (sha)",
                    stderr: String::from_utf8_lossy(&sha.stderr).into_owned(),
                });
            };
            git_sha
        }
    };
    Ok(XmlGitInfo {
        describe,
        commit_timestamp: timestamp,
        sha: git_sha,
    })
}

/// Write formatted codes generated with [`Module::generate_code`](crate::types::Module::generate_code).
pub fn write_to_files(
    files: HashMap<path::PathBuf, syn::File>,
    out_path: &path::Path,
) -> io::Result<()> {
    if !out_path.is_dir() {
        return Err(io::Error::from(io::ErrorKind::NotADirectory));
    }
    if fs::read_dir(out_path)?.next().is_some() {
        return Err(io::Error::new(
            io::ErrorKind::AlreadyExists,
            format!("out path `{}` is not empty", out_path.display()),
        ));
    }
    for (file_path, code) in files {
        fs::DirBuilder::new()
            .recursive(true)
            .create(out_path.join(&file_path).parent().unwrap())?;

        fs::write(out_path.join(&file_path), prettyplease::unparse(&code))?;
    }
    Ok(())
}
