//! Validate the address assignment and generate a flattened register map.

use thiserror::Error;

use crate::types::{Block, Module, ModuleBlockElements, Register};

#[derive(Debug, Clone)]
pub struct FlattenedRegisterEntry<'a> {
    /// contains all parents
    pub path: Vec<&'a str>,
    pub register: &'a Register,
    /// Id of multiple, if the register is multiple
    pub multiple_id: Option<u32>,
    /// Used bits
    pub field_masks: u32,
}

/// Flattened register map.
pub type FlattenedRegisterMap<'a> = Vec<Option<FlattenedRegisterEntry<'a>>>;

impl Module {
    /// Validate the address assignment, generating a flatten register map.
    /// Returns a tuple of [`FlattenedRegisterMap`] and [`ValidationError`]s vector.
    /// Empty errors vector means there is no errors in the register map.
    pub fn validate(&self) -> (FlattenedRegisterMap, Vec<ValidationError>) {
        let mut mapping = Vec::new();
        mapping.resize(self.size.try_into().unwrap(), None);
        let errors = Vec::new();
        self.fill(mapping, errors, (0x0, Vec::new()))
    }
}

#[derive(Debug, Error, PartialEq)]
pub enum ValidationError {
    #[error("address duplicated at 0x{address:08x}: \"{existing_name}\", \"{new_name}\"")]
    RegisterAddressDuplicated {
        address: u32,
        existing_name: String,
        new_name: String,
    },
    #[error("address 0x{address:08x} is out of range: \"{name}\"")]
    AddressOutofRange { address: u32, name: String },
    #[error(
        "field masks are duplicated at 0x{address:08x}: \"{reg_name}\", bits: 0x{duplicated_mask:08x}"
    )]
    FieldMaskDuplicated {
        address: u32,
        reg_name: String,
        duplicated_mask: u32,
    },
    #[error(
        "register \"{name}\" at 0x{address:08x} has mask 0x{reg_mask:08x} that doesn't match sum of field masks: 0x{fields_mask:08x}"
    )]
    RegisterAndFieldMaskMismatch {
        address: u32,
        name: String,
        reg_mask: u32,
        fields_mask: u32,
    },
    #[error("unsupported structure: {msg}")]
    UnsupportedStructure { msg: &'static str },
}

trait Validate {
    fn fill<'a>(
        &'a self,
        mapping: FlattenedRegisterMap<'a>,
        errors: Vec<ValidationError>,
        base: (u32, Vec<&'a str>),
    ) -> (FlattenedRegisterMap<'a>, Vec<ValidationError>);
}

impl Validate for Module {
    fn fill<'a>(
        &'a self,
        mut mapping: FlattenedRegisterMap<'a>,
        mut errors: Vec<ValidationError>,
        base: (u32, Vec<&'a str>),
    ) -> (FlattenedRegisterMap<'a>, Vec<ValidationError>) {
        for child in &self.elements_other {
            (mapping, errors) = child.fill(mapping, errors, base.clone());
        }
        (mapping, errors)
    }
}

impl Validate for ModuleBlockElements {
    fn fill<'a>(
        &'a self,
        mapping: FlattenedRegisterMap<'a>,
        errors: Vec<ValidationError>,
        base: (u32, Vec<&'a str>),
    ) -> (FlattenedRegisterMap<'a>, Vec<ValidationError>) {
        match self {
            ModuleBlockElements::Block(block) => block.fill(mapping, errors, base),
            ModuleBlockElements::Register(register) => register.fill(mapping, errors, base),
            ModuleBlockElements::Memory(_memory) => todo!(),
            ModuleBlockElements::Fifo(_fifo) => todo!(),
        }
    }
}

impl Validate for Block {
    fn fill<'a>(
        &'a self,
        mut mapping: FlattenedRegisterMap<'a>,
        mut errors: Vec<ValidationError>,
        base: (u32, Vec<&'a str>),
    ) -> (FlattenedRegisterMap<'a>, Vec<ValidationError>) {
        if self.multiple.is_some() {
            errors.push(ValidationError::UnsupportedStructure {
                msg: "multiple in block",
            });
        }

        let addr = base.0 + self.addr;
        let mut path = base.1;
        path.push(&self.name);
        for child in &self.elements {
            (mapping, errors) = child.fill(mapping, errors, (addr, path.clone()));
        }
        (mapping, errors)
    }
}

impl Validate for Register {
    fn fill<'a>(
        &'a self,
        mut mapping: FlattenedRegisterMap<'a>,
        mut errors: Vec<ValidationError>,
        base: (u32, Vec<&'a str>),
    ) -> (FlattenedRegisterMap<'a>, Vec<ValidationError>) {
        let addr = base.0 + self.addr;
        let path = base.1;

        let new_name = {
            let mut path = path.clone();
            path.push(&self.name);
            path.join("/")
        };
        let (reg_mask, duplicate_bits) = match self.elements.is_empty() {
            true => (0xffff_ffff, 0),
            false => self
                .elements
                .iter()
                .fold((0, 0), |(sum, _duplicate_bits), field| {
                    let field_mask = match &field.multiple {
                        Some(multiple) => (0..multiple.multiple)
                            .map(|multiple_id| field.mask << (multiple.offset * multiple_id))
                            .sum(),
                        None => field.mask,
                    };
                    (sum + field_mask, sum & field_mask)
                }),
        };
        if duplicate_bits != 0 {
            errors.push(ValidationError::FieldMaskDuplicated {
                address: addr,
                reg_name: new_name,
                duplicated_mask: duplicate_bits,
            });
            return (mapping, errors);
        }
        if let Some(mask) = self.mask {
            let has_fields = !self.elements.is_empty();
            if has_fields && mask != reg_mask {
                errors.push(ValidationError::RegisterAndFieldMaskMismatch {
                    address: addr,
                    name: new_name.clone(),
                    reg_mask: mask,
                    fields_mask: reg_mask,
                });
                return (mapping, errors);
            }
        };

        let (len, offset) = match &self.multiple {
            Some(multiple) => (multiple.multiple, multiple.offset),
            None => (1, 1),
        };
        for id in 0..len {
            let addr = addr + id * offset;
            let regmap: &mut Option<_> = match mapping.get_mut::<usize>(addr.try_into().unwrap()) {
                Some(regmap) => regmap,
                None => {
                    errors.push(ValidationError::AddressOutofRange {
                        address: addr,
                        name: self.name.clone(),
                    });
                    continue;
                }
            };
            if let Some(old) = regmap {
                let existing_name = {
                    let mut path = old.path.clone();
                    path.push(&old.register.name);
                    path.join("/")
                };

                if old.register.name != self.name {
                    errors.push(ValidationError::RegisterAddressDuplicated {
                        address: addr,
                        existing_name,
                        new_name: new_name.clone(),
                    });
                    continue;
                }
                log::trace!(
                    "defining same register at 0x{addr:08x}: {:?}, {:?}",
                    old.register,
                    self
                );
                let duplicate_bits = old.field_masks & reg_mask;
                if duplicate_bits != 0 {
                    errors.push(ValidationError::FieldMaskDuplicated {
                        address: addr,
                        reg_name: new_name.clone(),
                        duplicated_mask: duplicate_bits,
                    });
                }
                let multiple_id = if len == 1 { None } else { Some(id) };
                *regmap = Some(FlattenedRegisterEntry {
                    path: path.clone(),
                    register: self,
                    multiple_id,
                    field_masks: old.field_masks | reg_mask,
                })
            } else {
                let multiple_id = if len == 1 { None } else { Some(id) };
                *regmap = Some(FlattenedRegisterEntry {
                    path: path.clone(),
                    register: self,
                    multiple_id,
                    field_masks: reg_mask,
                })
            }
        }

        (mapping, errors)
    }
}

#[cfg(test)]
mod test {
    use itertools::Itertools;

    use crate::{
        types::{
            Field, Module, ModuleBlockElements, MultipleParams, Register, RwSpecifier, XmlGitInfo,
        },
        validator::ValidationError,
    };

    #[test]
    fn duplicate() {
        let reg1 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: None,
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![],
        };
        let reg2 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: None,
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![],
        };
        let module = Module {
            name: "module".to_string(),
            addr: 0,
            size: 0x10,
            amod: None,
            r#type: None,
            desc: None,
            elements_bitstring: vec![],
            elements_other: vec![
                ModuleBlockElements::Register(reg1),
                ModuleBlockElements::Register(reg2),
            ],
            xmlhash: [0; 32],
            git_info: XmlGitInfo {
                describe: "".to_string(),
                commit_timestamp: "".to_string(),
                sha: "".to_string(),
            },
        };

        let (maps, errors) = module.validate();
        eprintln!("{:#?}", maps);
        eprintln!(
            "{:#?}",
            maps.iter().filter(|e| e.is_some()).collect_vec().len()
        );
        assert_eq!(maps.len(), 0x10);
        assert!(!errors.is_empty());
        assert_eq!(maps.iter().filter(|e| e.is_some()).collect_vec().len(), 0x1);
    }

    #[test]
    fn duplicate_with_multiple() {
        let reg1 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: None,
            modf: RwSpecifier::R,
            multiple: Some(MultipleParams {
                multiple: 16,
                offset: 1,
            }),
            default: None,
            desc: None,
            elements: vec![],
        };
        let reg2 = Register {
            name: "test1".to_string(),
            addr: 15,
            r#type: crate::types::DataType::D32,
            mask: None,
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![],
        };
        let module = Module {
            name: "module".to_string(),
            addr: 0,
            size: 0x1000,
            amod: None,
            r#type: None,
            desc: None,
            elements_bitstring: vec![],
            elements_other: vec![
                ModuleBlockElements::Register(reg1),
                ModuleBlockElements::Register(reg2),
            ],
            xmlhash: [0; 32],
            git_info: XmlGitInfo {
                describe: "".to_string(),
                commit_timestamp: "".to_string(),
                sha: "".to_string(),
            },
        };

        let (maps, errors) = module.validate();
        assert_eq!(maps.len(), 0x1000);
        assert!(!errors.is_empty());
        assert_eq!(errors.len(), 1);
    }

    // Fields are orthogonal
    #[test]
    fn duplicate_reg_but_fields_are_ok() {
        let reg1 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: None,
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![
                Field {
                    name: "field_1".to_string(),
                    mask: 0x0000_00ff,
                    interface: None,
                    multiple: None,
                    default: None,
                    sclr: None,
                    desc: Some("field 1 description".to_string()),
                    elements: vec![],
                },
                Field {
                    name: "field_2".to_string(),
                    mask: 0x0000_ff00,
                    interface: None,
                    multiple: None,
                    default: None,
                    sclr: None,
                    desc: Some("field 2 description".to_string()),
                    elements: vec![],
                },
            ],
        };
        let reg2 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: None,
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![
                Field {
                    name: "field_3".to_string(),
                    mask: 0x00ff_0000,
                    interface: None,
                    multiple: None,
                    default: None,
                    sclr: None,
                    desc: Some("field 3 description".to_string()),
                    elements: vec![],
                },
                Field {
                    name: "field_4".to_string(),
                    mask: 0xff00_0000,
                    interface: None,
                    multiple: None,
                    default: None,
                    sclr: None,
                    desc: Some("field 4 description".to_string()),
                    elements: vec![],
                },
            ],
        };
        let module = Module {
            name: "module".to_string(),
            addr: 0,
            size: 0x10,
            amod: None,
            r#type: None,
            desc: None,
            elements_bitstring: vec![],
            elements_other: vec![
                ModuleBlockElements::Register(reg1),
                ModuleBlockElements::Register(reg2),
            ],
            xmlhash: [0; 32],
            git_info: XmlGitInfo {
                describe: "".to_string(),
                commit_timestamp: "".to_string(),
                sha: "".to_string(),
            },
        };

        let (maps, errors) = module.validate();
        eprintln!("{:#?}", maps);
        eprintln!(
            "{:#?}",
            maps.iter().filter(|e| e.is_some()).collect_vec().len()
        );
        assert_eq!(maps.len(), 0x10);
        assert!(errors.is_empty());
        assert_eq!(maps.iter().filter(|e| e.is_some()).collect_vec().len(), 0x1);
    }

    #[test]
    fn register_with_mask_and_field() {
        let reg1 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: Some(0x000f_00ff),
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![
                Field {
                    name: "field_1".to_string(),
                    mask: 0x0000_0001,
                    interface: None,
                    multiple: Some(MultipleParams {
                        multiple: 8,
                        offset: 1,
                    }),
                    default: None,
                    sclr: None,
                    desc: Some("field 1 description".to_string()),
                    elements: vec![],
                },
                Field {
                    name: "field_2".to_string(),
                    mask: 0x0003_0000,
                    interface: None,
                    multiple: Some(MultipleParams {
                        multiple: 2,
                        offset: 2,
                    }),
                    default: None,
                    sclr: None,
                    desc: Some("field 1 description".to_string()),
                    elements: vec![],
                },
            ],
        };
        let module = Module {
            name: "module".to_string(),
            addr: 0,
            size: 0x1,
            amod: None,
            r#type: None,
            desc: None,
            elements_bitstring: vec![],
            elements_other: vec![ModuleBlockElements::Register(reg1)],
            xmlhash: [0; 32],
            git_info: XmlGitInfo {
                describe: "".to_string(),
                commit_timestamp: "".to_string(),
                sha: "".to_string(),
            },
        };

        let (maps, errors) = module.validate();
        eprintln!("{:#?}", maps);
        eprintln!("{:#?}", errors);
        eprintln!(
            "{:#?}",
            maps.iter().filter(|e| e.is_some()).collect_vec().len()
        );
        assert_eq!(maps.len(), 0x1);
        assert!(errors.is_empty());
        assert_eq!(maps.iter().filter(|e| e.is_some()).collect_vec().len(), 0x1);
    }

    #[test]
    fn register_with_invalid_mask_and_field() {
        let reg1 = Register {
            name: "test1".to_string(),
            addr: 0,
            r#type: crate::types::DataType::D32,
            mask: Some(0x0000_000f),
            modf: RwSpecifier::R,
            multiple: None,
            default: None,
            desc: None,
            elements: vec![Field {
                name: "field_1".to_string(),
                mask: 0x0000_0001,
                interface: None,
                multiple: Some(MultipleParams {
                    multiple: 8,
                    offset: 1,
                }),
                default: None,
                sclr: None,
                desc: Some("field 1 description".to_string()),
                elements: vec![],
            }],
        };
        let module = Module {
            name: "module".to_string(),
            addr: 0,
            size: 0x1,
            amod: None,
            r#type: None,
            desc: None,
            elements_bitstring: vec![],
            elements_other: vec![ModuleBlockElements::Register(reg1)],
            xmlhash: [0; 32],
            git_info: XmlGitInfo {
                describe: "".to_string(),
                commit_timestamp: "".to_string(),
                sha: "".to_string(),
            },
        };

        let (maps, errors) = module.validate();
        eprintln!("{:#?}", maps);
        eprintln!("{:#?}", errors);
        eprintln!(
            "{:#?}",
            maps.iter().filter(|e| e.is_some()).collect_vec().len()
        );
        assert_eq!(maps.len(), 0x1);
        assert_eq!(errors.len(), 1);
        assert_eq!(maps.iter().filter(|e| e.is_some()).collect_vec().len(), 0x0);
        assert_eq!(
            errors.first().unwrap(),
            &ValidationError::RegisterAndFieldMaskMismatch {
                address: 0x0,
                name: "test1".to_string(),
                reg_mask: 0x0000_000f,
                fields_mask: 0x0000_00ff
            }
        );
    }
}
